/**
 * \file
 *
 * \brief CDC Application Main functions
 *
 * Copyright (c) 2014-2018 Microchip Technology Inc. and its subsidiaries.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Subject to your compliance with these terms, you may use Microchip
 * software and any derivatives exclusively with Microchip products.
 * It is your responsibility to comply with third party license terms applicable
 * to your use of third party software (including open source software) that
 * may accompany Microchip software.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE,
 * INCLUDING ANY IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY,
 * AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE
 * LIABLE FOR ANY INDIRECT, SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL
 * LOSS, DAMAGE, COST OR EXPENSE OF ANY KIND WHATSOEVER RELATED TO THE
 * SOFTWARE, HOWEVER CAUSED, EVEN IF MICROCHIP HAS BEEN ADVISED OF THE
 * POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.  TO THE FULLEST EXTENT
 * ALLOWED BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY
 * RELATED TO THIS SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY,
 * THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
 *
 * \asf_license_stop
 *
 * History:
 * - 18/03/2020 Quentin Bolsee
 *      Created a USB-UART full duplex asynchronous tunnel
 *
 */
/*
 * Support and FAQ: visit <a href="https://www.microchip.com/support/">Microchip Support</a>
 */

#include <asf.h>
#include <string.h>
#include "conf_usb.h"
#include "usb_protocol_cdc.h"


/* Buffer size for USART receiver array */
// must be ^2
#define USART_DATA_LENGTH 128
#define INCREMENT_RING(x) x=(x+1)&(USART_DATA_LENGTH-1)

/* buffers */
//static uint8_t usb_read_buffer[USB_DATA_LENGTH];

static uint8_t rx_buffer[USART_DATA_LENGTH];
static iram_size_t rx_read_ptr = 0; // main loop reads here
static iram_size_t rx_write_ptr = 0; // ISR writes here
static bool rx_buffer_full = false;

static uint8_t tx_buffer[USART_DATA_LENGTH];
static iram_size_t tx_read_ptr = 0; // ISR reads here
static iram_size_t tx_write_ptr = 0; // main loop writes here
static bool tx_buffer_full = false;


/* usart parameters */
static struct usart_module usart_2;


static volatile bool main_b_cdc_enable = false;


static void usart_read_callback(const struct usart_module *const module) {
	bool can_write = true;

	// NB: inside of an ISR, no need to enter a critical section
	INCREMENT_RING(rx_write_ptr);
	if (rx_write_ptr == rx_read_ptr) {
		rx_buffer_full = true;
		can_write = false;
	}
	
	if (can_write) {
		usart_read_buffer_job(&usart_2, (uint8_t *)rx_buffer+rx_write_ptr, 1);
	}
}

static void usart_write_callback(const struct usart_module *const module) {
	bool can_write = false;

	// NB: inside of an ISR, no need to enter a critical section
	INCREMENT_RING(tx_read_ptr);
	tx_buffer_full = false;
	if (tx_read_ptr != tx_write_ptr) {
		can_write = true;
	}

	if (can_write) {
		usart_write_buffer_job(&usart_2, (uint8_t *)tx_buffer+tx_read_ptr, 1);
	}
}

static void configure_usart(void) {
	struct usart_config config_usart;
		
	usart_get_config_defaults(&config_usart);

	// UART CONFIG
	config_usart.baudrate    = 115200;
	config_usart.stopbits    = USART_STOPBITS_2;
	config_usart.parity      = USART_PARITY_EVEN;
	config_usart.mux_setting = USART_RX_3_TX_2_XCK_3;
	config_usart.pinmux_pad0 = PINMUX_UNUSED;
	config_usart.pinmux_pad1 = PINMUX_UNUSED;
	config_usart.pinmux_pad2 = PINMUX_PA04C_SERCOM0_PAD2;
	config_usart.pinmux_pad3 = PINMUX_PA05C_SERCOM0_PAD3;
	config_usart.start_frame_detection_enable = true;
	
	while (usart_init(&usart_2, SERCOM0, &config_usart) != STATUS_OK) {
	}

	usart_enable(&usart_2);
	
	//usart_register_callback()
	usart_register_callback(&usart_2, (usart_callback_t)usart_read_callback, USART_CALLBACK_BUFFER_RECEIVED);
	usart_enable_callback(&usart_2, USART_CALLBACK_BUFFER_RECEIVED);
	usart_register_callback(&usart_2, (usart_callback_t)usart_write_callback, USART_CALLBACK_BUFFER_TRANSMITTED);
	usart_enable_callback(&usart_2, USART_CALLBACK_BUFFER_TRANSMITTED);
	
	usart_read_buffer_job(&usart_2, (uint8_t *)rx_buffer+rx_write_ptr, 1);
}

/*! \brief Main function. Execution starts here.
 */
int main(void) {
	//size_t available_cdc;
	uint8_t c;
	bool tx_ready;
	bool rx_ready;
	
	irq_initialize_vectors();
	cpu_irq_enable();
	
	// Initialize the sleep manager
	sleepmgr_init();

	system_init();

	// Start USB stack to authorize VBus monitoring
	udc_start();

	configure_usart();

	// The main loop manages only the power mode
	// because the USB management is done by interrupt
	while (true) {
		sleepmgr_enter_sleep();
		if (!main_b_cdc_enable) {
			continue;
		}
		
		// TX logic
		if (udi_cdc_is_rx_ready()) {
			tx_ready = false;
			//udi_cdc_putc(udi_cdc_getc());
			
			//available_cdc = udi_cdc_read_no_polling(usb_read_buffer, (iram_size_t)USB_DATA_LENGTH);
			//udi_
			c = udi_cdc_getc();
			
			system_interrupt_enter_critical_section();
			if (!tx_buffer_full) {
				if (tx_read_ptr == tx_write_ptr) {
					tx_ready = true;
				}
				tx_buffer[tx_write_ptr] = c;
				INCREMENT_RING(tx_write_ptr);
				if (tx_read_ptr == tx_write_ptr) {
					tx_buffer_full = true;
				}
			}
			system_interrupt_leave_critical_section();
			
			if (tx_ready) {
				usart_write_buffer_job(&usart_2, (uint8_t *)tx_buffer+tx_read_ptr, 1);
			}
			
			//udi_cdc_putc(c);
			//udi_cdc_read_buf(usb_read_buffer, (iram_size_t)USB_DATA_LENGTH);
			//udi_cdc_write_buf(usart_2_read_buffer, (iram_size_t)available_cdc);
			//usart_write_buffer_job(&usart_2, (uint8_t *)usb_read_buffer, available_cdc);
			
		}
		
		// RX logic
		rx_ready = false;
		system_interrupt_enter_critical_section();
		if (rx_buffer_full) {
			// buffer full, so data is available
			rx_buffer_full = false;
			rx_ready = true;
		} else if (rx_read_ptr == rx_write_ptr) {
			// buffer emtpy, nothing to write
			system_interrupt_leave_critical_section();
			continue;
		}
		system_interrupt_leave_critical_section();
		
		if (rx_ready) {
			usart_read_buffer_job(&usart_2, (uint8_t *)rx_buffer+rx_write_ptr, 1);
		}
		c = rx_buffer[rx_read_ptr];
		INCREMENT_RING(rx_read_ptr);
		udi_cdc_putc(c);
	}
}

bool main_cdc_enable(uint8_t port) {
	main_b_cdc_enable = true;
	return true;
}

void main_cdc_disable(uint8_t port) {
	main_b_cdc_enable = false;
}
